<?php

namespace Wienkit\BeslistShopitemClient;

use Wienkit\BeslistShopitemClient\Entities\AuthenticationResult;
use Wienkit\BeslistShopitemClient\Exceptions\BeslistShopitemException;
use Wienkit\BeslistShopitemClient\Entities\ShopItem;

class BeslistShopitemClientKoongo extends BeslistShopitemClient
{
    const BASE_URL = 'https://shopitem.api.beslist.nl/';
    const TEST_URL = 'https://test-shopitem.api.beslist.nl/';
//    https://test-shopitem.api.beslist.nl/swagger
    /**
     * The Query to run against the FileSystem
     * @var \RestClient;
     */
    private $api;


    /**
     * BeslistShopitemClient constructor.
     * @param $apiKey
     */
    public function __construct($apiKey)
    {
        $this->api = new \RestClient([
            'base_url' => self::BASE_URL,
            'headers' => ['apikey' => $apiKey],
        ]);
        $this->api->register_decoder('json', function($data){
            return json_decode($data, TRUE);
        });
    }

    /**
     * Enable or disable testmode (default disabled)
     * @param bool $testMode
     */
    public function setTestMode($testMode)
    {
        if($testMode) {
            $this->api->set_option('base_url', self::TEST_URL);
        } else {
            $this->api->set_option('base_url', self::BASE_URL);
        }
    }


    /**
     * Authenticate with the Beslist Shopitem API
     * @return AuthenticationResult
     * @throws BeslistShopitemException
     */
    public function authenticate()
    {
        try {
            $result = $this->api->get("auth/v1/shops");
            if($result->info->http_code == 200) {
                return new AuthenticationResult($result->decode_response());
            } else {
                throw new BeslistShopitemException(null, $result);
            }
        } catch (\RestClientException $exception) {
            throw new BeslistShopitemException($exception);
        }
    }
    
    public function getSites() {
        
        $result = $this->authenticate();
        if( !$result->sites || !count( $result->sites)) {
            throw new BeslistShopitemException( 'No shop included in API');
        }
        return $result->sites;
    }
    
    /**
     * @param $shopId
     * @param $productId
     * @param ShopItem $shopItem
     * @return boolean success
     * @throws BeslistShopitemException
     */
    public function updateShopItem($shopId, $productId, ShopItem $shopItem)
    {
    	try {
    		$body = json_encode($shopItem);
    		$result = $this->api->put(self::API_VERSION.self::API_PREFIX.$shopId . '/offers/' . $productId, $body);
    		if ($result->info->http_code == 200) {
    			return true;
    		} else {
    			throw new BeslistShopitemException(NULL, $result);
    		}
    	} catch (\RestClientException $exception) {
    		throw new BeslistShopitemException($exception);
    	}
    }

    /**
     * Retrieve a ShopItem
     * @param $shopId
     * @param $productId
     * @return ShopItem
     * @throws BeslistShopitemException
     */
    public function getShopItem($shopId, $productId)
    {
    	try {
    		$result = $this->api->get(self::API_VERSION.self::API_PREFIX.$shopId . '/offers/' . $productId);
    		if($result->info->http_code == 200) {
    			return ShopItem::fromArray($result->decode_response());
    		} else {
    			throw new BeslistShopitemException(NULL, $result);
    		}
    	} catch (\RestClientException $exception) {
    		throw new BeslistShopitemException($exception);
    	}
    }    
    
    public function updateShopItems( $shopId, $items) {        
        try {        	
            $result = $this->api->put(self::API_VERSION.self::API_PREFIX.$shopId.'/batch', json_encode($items));
                       
            if($result->info->http_code == 201) {            
                $response = array('url' => $result->headers->location);
                if( !isset( $response['url']) || empty($response['url'])) {
                    throw new \Exception( "Wrong response for updateShopItems. Parameter 'url' is missing");
                }
                return $response;
            } else if($result->info->http_code == 400) {
            	$response = $result->decode_response();
            	if(isset($response['error']))
            	{
            		$message = $response['error'];
            		if(isset($response['errorDetails']))
            			$message .= " Details".": ".json_encode($response['errorDetails']);            			
            		throw new \Exception($message);
            	}
            	else
            		throw new BeslistShopitemException(NULL, $result);
            } else {            	
                throw new BeslistShopitemException(NULL, $result);
            }
        } catch (\RestClientException $exception) {
            throw new BeslistShopitemException($exception);
        }
    }
    
    public function getBatchResult( $shopId, $batchId) {
        
        try {
            $result = $this->api->get(self::API_VERSION.self::API_PREFIX.$shopId.'/batch/' . $batchId);
            if($result->info->http_code == 200) {
                
                $response = $result->decode_response();
                if( !isset( $response['batch'])) {
                    throw new \Exception( "Wrong response for getBatchResult. Parameter 'batch' is missing");
                }
                return $response;
                
            } else {
                throw new BeslistShopitemException(NULL, $result);
            }
        } catch (\RestClientException $exception) {
            throw new BeslistShopitemException($exception);
        }
    }
}
